/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: SendErrr.cpp$
 * $Date: Sun Aug 10 06:36:50 1997$
 *
 Description:
	Sends canned error messages for specific status codes.

\*____________________________________________________________________________*/
//$SourceTop:$

#include <ctype.h>

#include "HandBase.h"
#include "HTTPCore.h"
#include "PIStrStr.h"
#include "DblList.h"
#include "StrToken.h"

/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class ErrorMap
{
public:
	int iStatus;
	PIString sErrorFile;

	ErrorMap( int iTheStatus, const char *pErrorFile )
	: iStatus( iTheStatus ), sErrorFile( pErrorFile )
		{};
};

/*____________________________________________________________________________*\
 *
 Description:
\*____________________________________________________________________________*/
#define KEY_CONF_IGNORESTATUS		"IgnoreStatus"
#define KEY_CONF_DEFAULTMESSAGE		"DefaultMessage"
#if 0
	/*
	** HTML documentation for this handler
	*/
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	SendErrorMessage

Description:
	This handler object sends specific canned error message text for each
	HTTP error status code.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>IgnoreStatus
<TD>-
<TD>0, 200, etc.
<TD>HTTP response codes
<TD>IgnoreStatus="0 200"

<TR>
<TD>&ltnnn&gt
<TD>-
<TD>Fragments/404.html, msgs/403.html, etc.
<TD>Response specific error messages
<TD>404="Fragments/404.html"; 403="msgs/403.html"

<TR>
<TD>DefaultMessage
<TD>+
<TD>Fragments/Unknown.html, mgs/unknown.html, etc.
<TD>Default error message
<TD>DefaultMessage="Framents/Unknown.html"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5>
	IgnoreStatus
</H5>
HTTP response codes which should not be considered errors. This field
	takes the form of a sequence of 3 digit decimal codes seperated by
	spaces.
	<STRONG>NOTE:</STRONG> Typically the values 0 and 200 will always be 
	included.
<H5>
	&ltnnn&gt
</H5>
This directive is used to specify the error message file to be sent for a 
	specific status code.

<H5>
	DefaultMessage
</H5>
The default error message. This is used for status codes which have
	not been specified as ignorable, but which have not been explicitly 
	defined.

Phase:
	HANDLE

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_CONTINUE if no action taken.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>
	&ltObject&gt
		Name SendErrorMessage
		Class SendErrorMessageClass
		IgnoreStatus "0 200"
	&lt/Object&gt

	&ltObject&gt
		...
		Handle SendErrorMessage DefaultMessage="Fragments\Unknown.html" 403="Fragments\Forbidden.html"
		...
	&lt/Object&gt
	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif
/*____________________________________________________________________________*\
 *
 Class:
 Description:
\*____________________________________________________________________________*/
class SendErrorMessage : public HandlerBaseHTTP
{
private:
	/* --- forbid copy constructor --- */
	SendErrorMessage( const SendErrorMessage &t )
	: HandlerBaseHTTP( t )
		{ assert( 0 ); };

	/* ---
	Configuration data
	--- */
	DblList lIgnoreCodes;
	DblList lErrorMaps;
	PIString sDefaultFile;

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		PIOStrStream os;
		os << pWhere << "SendErrorMessage: ";
		if ( !PIUtil_stricmp( KEY_CONF_IGNORESTATUS, pVariable ) )
			{
			StringTokenizer tTokens( pValue, " " );
			for( int i=0; i<tTokens.NumTokens(); i++)
				{
				lIgnoreCodes.Append( (DblList::type)
					atoi( tTokens.GetToken( i ) ) );
				};
			}
		else if ( !PIUtil_stricmp( KEY_CONF_DEFAULTMESSAGE, pVariable ) )
			{
#if 0
			Pi3String *pString = Pi3String_new( pValue );
			HTTPCore_relativeToAbsolutePath( pValue, pString );
			sDefaultFile = Pi3String_getPtr( pString );
			Pi3String_delete( pString );
#endif
            sDefaultFile = pValue;
			}
		else if ( isdigit( *pVariable ) )
			{
#if 0
			int iTmp = atoi( pVariable );
			Pi3String *pTmp = Pi3String_new( pValue );
			HTTPCore_relativeToAbsolutePath( pValue, pTmp );
			PIString sTmp( Pi3String_getPtr( pTmp ) );
			Pi3String_delete( pTmp );
			lErrorMaps.Append( (DblList::type) 
				PI_NEW( ErrorMap( iTmp, sTmp ) ) ); 
#endif
			int iTmp = atoi( pVariable );
			lErrorMaps.Append( (DblList::type) 
				PI_NEW( ErrorMap( iTmp, pValue ) ) ); 
			}
		else
			{
			os << "Unknown directive '" << pVariable <<
				"'" << ends;
			CONFIG_ERR( Object(), os.str() );
			return 0;
			};

		return 1;
		};

public:
	SendErrorMessage( PIObject *pObject, int iArgc, const char *ppArgv[] )
	:	HandlerBaseHTTP( pObject )
		{
		ReadParameters( iArgc, ppArgv );

		/* --- 
		Make sure there's a default	
		--- */
		if ( sDefaultFile==PIString::Empty() )
			{
			CONFIG_ERR( Object(), "SendErrorMessage: 'Default' not defined" );
			SetOK( 0 );
			return;
			};
		};

	virtual ~SendErrorMessage()
		{
		for( DblListIterator i( lErrorMaps ); !i.BadIndex(); i++)
			{
			PI_DELETE( (ErrorMap *)i.Current() );
			};
		};

	int Handle( int iPhase, PIHTTP &tPIHTTP, PIIOBuffer &tB )
		{
		if ( iPhase!=PH_HANDLE ) { return PIAPI_ERROR; };

		PIDB *pR = tPIHTTP.pResponseDB;

		int iStatus = tPIHTTP.iStatus;

		/* ---
		See is this a status code that we should ignore
		--- */
		for( DblListIterator i( lIgnoreCodes ); !i.BadIndex(); i++ )
			{
			if ( iStatus==(int)i.Current() )	
				{ return PIAPI_CONTINUE; };
			};

		/* ---
		Find the mapping with the appropriate status
		--- */
		const char *pErrorFile = 0;
		for( DblListIterator j( lErrorMaps ); !j.BadIndex(); j++ )
			{
			ErrorMap *pMap = (ErrorMap *)j.Current();
			assert( pMap );
			if ( iStatus==pMap->iStatus )	
				{
				pErrorFile = pMap->sErrorFile;
				break;
				};
			};

		/* ---
		Assign default if necessary
		--- */
		if ( !pErrorFile )
			{
			pErrorFile = sDefaultFile;
			};

        /* ---
        Change from relative to absolute path
        --- */
        Pi3String *pString = Pi3String_new( pErrorFile );
        HTTPCore_relativeToAbsolutePath( tPIHTTP.pHostDB, pErrorFile,
            pString );
        PIString sErrorFile( Pi3String_getPtr( pString ) );
        Pi3String_delete( pString );
        pErrorFile = sErrorFile;

		/* ---
		Memory map the file and then send it
		--- */
		PIFInfo *pFileInfo = HTTPCore_getCachedFile( pErrorFile );
		if ( !pFileInfo )
			{ return PIAPI_ERROR; };

		const char *pMap = 0;
		int iLen = 0;

		/* --- get filemap object --- */
		PIFMap *pFileMap = PIFMap_new( pFileInfo );
		if ( !pFileMap )
			{ goto sendfile_error; };

		/* ---
		do response
		--- */
		if (	HTTPCore_sendGeneralHeaders( &tPIHTTP ) ||
				HTTPCore_sendEntityHeaders( &tPIHTTP, pR ) )
			{ goto sendfile_error; };

		/* --- send mapped file --- */
		pMap = (const char *)PIFMap_lock( pFileMap, &iLen );

		if ( !pMap )
			{ goto sendfile_error; };
		
		if ( pMap )
			{
			PIIOBuffer_write( &tB, pMap, iLen, PIIOBUF_NONE ); 
			}; 

		PIFMap_delete( pFileMap );
		HTTPCore_releaseCachedFile( pFileInfo );
		return PIAPI_COMPLETED;

sendfile_error:
		if ( pFileMap )
			{ PIFMap_delete( pFileMap ); };
		HTTPCore_releaseCachedFile( pFileInfo );
		tPIHTTP.iStatus = ST_INTERNALERROR;
		return PIAPI_ERROR;
		};
};

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int SendErrorMessage_constructor( PIObject *pObj,
	int iArgc, const char *ppArgv[] )
{
	return HandlerBaseHTTP_constructor( pObj, PI_NEW( SendErrorMessage( pObj,
		iArgc, ppArgv ) ) );
}

#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name SendErrorMessageClass
		Type LogicExtension
		Library HTTP
		OnClassLoad HandlerBaseHTTP_onClassLoad
		Constructor SendErrorMessage_constructor
		CopyConstructor HandlerBaseHTTP_copyConstructor
		Destructor HandlerBaseHTTP_destructor
		Execute HandlerBaseHTTP_execute
	</Class>

	<Object>
		Name SendErrorMessage
		Class SendErrorMessageClass
	</Object>

/*___+++CNF_END+++___*/
#endif

