/*____________________________________________________________________________*\
 *

 Copyright (c) 1997 John Roy. All rights reserved.

 These sources, libraries and applications are
 FREE FOR COMMERCIAL AND NON-COMMERCIAL USE
 as long as the following conditions are adhered to.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.

 3. Redistributions of any form whatsoever and all advertising materials 
    mentioning features must contain the following
    acknowledgment:
    "This product includes software developed by John Roy
    (http://www.johnroy.com/pi3/)."

 THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHORS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
 DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
 GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 OF THE POSSIBILITY OF SUCH DAMAGE.

 *____________________________________________________________________________*|
 *
 * $Source: UDPIPSck.cpp$
 * $Date: Sun Aug 10 06:35:53 1997$
 *
 Description:
	Implements a UDP/IP socket class

\*____________________________________________________________________________*/
/* $SourceTop:$ */

/* --- ANSI --- */
#include <iostream.h>
#include <fstream.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <errno.h>

/* --- system includes --- */
#if POSIX
#	define USE_FLOCK		/* what OS's need this? */
#	define _BSD_COMPAT		/* this is for IRIX 5.3 */
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <sys/file.h>
#	include <fcntl.h>
#	include <netinet/in.h>
#	include <netinet/tcp.h>
#	include <netdb.h>
#	include <arpa/inet.h>
#	include <unistd.h>
#	define closesocket(x)	close(x)
#	define SOCKET_ERROR		(-1)

/*
 * Prototype gethostname
 */
int gethostname( char *, int );
#elif WIN32
#	include <winsock.h>
#else
#	error Unsupported Operating system
#endif

#include "Pi2API.h"

/*
** Macros
*/
#define LISTEN_QUEUE_SIZE				20
#define DEFAULT_TIMEOUT					360

/* --- Values from configuration file --- */
#define KEY_CONF_BINDPORT				"BindPort"
#define KEY_CONF_BINDHOST				"BindHost"
#define KEY_CONF_ACCEPTTIMEOUT			"AcceptTimeout"
#define KEY_CONF_RECVTIMEOUT			"RecvTimeout"
#define KEY_CONF_SENDTIMEOUT			"SendTimeout"
#define KEY_CONF_PEERADDRESSVARIABLE	"PeerIPVariable"
#define KEY_CONF_PEERHOSTNAMEVARIABLE	"PeerHostNameVariable"
#define KEY_CONF_LOCALADDRESSVARIABLE	"LocalAddressVariable"
#define KEY_CONF_ACCEPTLOCKFILE			"AcceptLockFile"
#define KEY_CONF_FLAGS					"Flags"

/* --- Values from configuration tree --- */
#define KEY_CONF_TYPE					"Type"
#define VALUE_PASSIVE					"Passive"
#define VALUE_ACTIVE					"Active"

/*
#define D	{ cerr << __FILE__ << ", " << __LINE__ << endl; };
*/
#define D  

#define STDERR (cerr << "UDPIPIO: ")

/*____________________________________________________________________________*\
 *
 Description:
	Messages
\*____________________________________________________________________________*/
const char szLockFileMsg[]="\n\
** ---\n\
**\n\
** The lockfile is used for synchronization amoungst multiple processes \n\
** using the same UDP/IP socket.\n\
**\n\
** Either remove the 'AcceptLockFile' directive and the named lockfile\n\
** or change the configuration to ensure that the lockfile is created.\n\
**\n\
** ---\
";
#if POSIX
const char szBindFailedAccess[]="\n\
** ---\n\
**\n\
** Failed to bind to the specified UDP/IP port because of insufficient \n\
** privileges. \n\
** \n\
** Note that use of UDP/IP ports 0-1023 is restricted on UNIX systems \n\
** (server must run initially as root).\n\
** \n\
** ---\
";
#else
const char szBindFailedAccess[]="\n\
** ---\n\
**\n\
** Failed to bind to the specified UDP/IP port because of insufficient \n\
** privilages. \n\
** \n\
** ---\
";
#endif
const char szBindFailedInUse[]="\n\
** ---\n\
**\n\
** Failed to bind to the specified UDP/IP port because the port \n\
** is already in use. \n\
** \n\
** Another process may be currently using the specified port.\n\
** \n\
** ---\
";

/*____________________________________________________________________________*\
 *
 Description:
	Functions used internally.
\*____________________________________________________________________________*/
class MainSocketData;
class UDPIPSocketData;
int Internal_initActiveSocket( PIObject *pObject );
int Internal_initPassiveSocket( PIObject *pObject,
		UDPIPSocketData *pProtoType );

/*____________________________________________________________________________*\
 *
 Description:
	Fast keys for faster access to variables
\*____________________________________________________________________________*/
static const char *pFKSendTimeout = 0;
static const char *pFKRecvTimeout = 0;

/*____________________________________________________________________________*\
 *
 Description:
		Configuration flags
\*____________________________________________________________________________*/
#define VALUE_FLAG_OWNDB				"OwnDB"
#define VALUE_FLAG_SETREMOTEADDRESS		"SetPeerAddress"
#define VALUE_FLAG_DNSREVERSELOOKUP		"DNSReverseLookup"
#define VALUE_FLAG_SETLOCALADDRESS		"SetLocalAddress"

#define FLG_OWNDB						0x0001
#define FLG_SETREMOTEADDRESS			0x0002
#define FLG_DNSREVERSELOOKUP			0x0004
#define FLG_SETLOCALADDRESS				0x0008
#define FLG_ERROR						0x8000

#define MAX_PACKET_SIZE                 1024

/*____________________________________________________________________________*\
** Class for PIDBTYPE, Variable name and PIDBFLAG
\*____________________________________________________________________________*/
class DBInfo
{
public:
	int iType;
	const char *pFKVariableName;
	int iFlags;

	DBInfo( const char *pDefinition )
	:	iType( -1 ), pFKVariableName( 0 ), iFlags( 0 )
		{
		if ( !pDefinition )
				{ return; };
		/* ---

		Decode something like the following:-

				STRING:PeerIP:0

		--- */
		enum { BUF_SIZE=2047 };
		char szBuf[BUF_SIZE+1];
		int i=0;
		const char *pS = pDefinition;
		for(; i<BUF_SIZE && pS[i] && pS[i]!=':'; i++) { szBuf[i]=pS[i]; };
		szBuf[i] = '\0';

		iType = PIDB_pidbtypeNameToNumber( szBuf );

		if ( i<BUF_SIZE && pS[i] ) { i++; };
		pS = &(pS[i]); i=0;
		for(; i<BUF_SIZE && pS[i] && pS[i]!=':'; i++) { szBuf[i]=pS[i]; };
		szBuf[i] = '\0';

		pFKVariableName = PIDB_getFastKey( pS, iType );

		if ( i<BUF_SIZE && pS[i] ) { i++; };
		pS = &(pS[i]); i=0;
		for(; i<BUF_SIZE && pS[i] && pS[i]!=':'; i++) { szBuf[i]=pS[i]; };
		szBuf[i] = '\0';

		if ( *szBuf )
				{ iFlags = PIDB_pidbflagsNameToNumber( szBuf ); };
		};

	~DBInfo()
		{
		};

	inline int IsOK()			   { return pFKVariableName && iType>=0; };
};

/*____________________________________________________________________________*\
 *
 Description:
	Documentation for this IO object.
\*____________________________________________________________________________*/

#if 0
/*___+++HTMLDOC_BEGIN+++___*/
Name:
	UDPIPIO

Description:
	This object encapsulates UDP/IP based IO communication.

Options:
<H5>Overview</H5>
<TABLE BORDER=1>
<TH>Option
<TH>Default
<TH>Values
<TH>Short Description
<TH>Example(s)

<TR>
<TD>Type
<TD>Passive
<TD>"Passive", "Active"
<TD>A Pi3 object
<TD>Type "Passive"

<TR>
<TD>BindPort
<TD>+ (passive only)
<TD>1..n
<TD>local port
<TD>BindPort 8080

<TR>
<TD>BindHost
<TD>+ (passive only)
<TD>nnn.nnn.nnn.nnn or host.org.dom
<TD>local host
<TD>BindHost "localhost"

<TR>
<TD>AcceptTimeout
<TD>360
<TD>n seconds or -1 (infinite)
<TD>Timeout on accept
<TD>AcceptTimeout -1

<TR>
<TD>RecvTimeout
<TD>-
<TD>n seconds or -1 (infinite)
<TD>Timeout on receive
<TD>RecvTimeout 45

<TR>
<TD>SendTimeout
<TD>-
<TD>n seconds or -1 (infinite)
<TD>Timeout on send
<TD>SendTimeout 30

<TR>
<TD>PeerAddressVariable
<TD>-
<TD>&ltdb type&gt:&ltname&gt
<TD>Variable to hold Peer IP address
<TD>PeerAddressVariable "STRING:RemoteAddr"

<TR>
<TD>PeerHostNameVariable
<TD>-
<TD>&ltdb type&gt:&ltname&gt
<TD>Variable to hold Peer hostname
<TD>PeerHostNameVariable "STRING:RemoteHost"

<TR>
<TD>LocalAddressVariable
<TD>-
<TD>&ltdb type&gt:&ltname&gt
<TD>Variable to hold local IP address
<TD>LocalAddressVariable "STRING:LocalAddr"

<TR>
<TD>AcceptLockFile
<TD>-
<TD>&ltfilename&gt
<TD>File containing lock filename
<TD>AcceptLockFile "Logs/lockfile.txt"

<TR>
<TD>Flags
<TD>0 (no flags set)
<TD>"OwnDB" | "SetPeerAddress" | "DNSReverseLookup"
<TD>Flags which effect operation
<TD>Flags "OwnDB|SetPeerAddress|DNSReverseLookup"

</TABLE>
<STRONG>-</STRONG> in the <IT>default</IT> indicates no default<BR>
<STRONG>+</STRONG> in the <IT>default</IT> indicates the field is mandatory<BR>

<H4>Description of Options</H4>
<H5> Type </H5>
Specify whether child objects the prototype IO object will be an active (client)
or passive (server) socket.

<H5> BindPort </H5>
The local port to listen on for passive sockets.

<H5> BindHost </H5>
The local IP address or hostname of the network interface to listen on
passive sockets.

<H5> AcceptTimeout </H5>
Accept timeout for passive sockets, a value of -1 specifies
no timeout.

<H5> RecvTimeout </H5>
Timeout on receiving data for active and passive sockets. A value
of -1 specifies no timeout.

<H5> SendTimeout </H5>
Timeout on sending data for active and passive sockets. A value
of -1 specifies no timeout.

<H5> PeerAddressVariable </H5>
The variable name with which to set the ip address of the remote
peer if the 'SetPeerAddress' flag is set.

<H5> PeerHostNameVariable </H5>
The variable name with which to set the hostname of the remote
peer if the 'SetPeerAddress' and 'DNSReverseLookup' flags are set.

<H5> LocalAddressVariable </H5>
The variable name with which to set the local IP address of the socket if
the 'SetLocalAddress' flag is set.

<H5> AcceptLockFile </H5>
The path to a file containing the lockfile for this server. The
lockfile is used to synchronization accepting of new connection
amoungst a group of processes listening for connections on the
same file descriptor. Note this is only applicable to POSIX operating
systems where the <CODE>fork()</CODE> system call may have been invoked
to spawn a single process with a single listening file descriptor into
multiple processes with multiple copies of the same file descriptor.

<H5> Flags </H5>
This options allows various options to be enabled and disabled. The
value of the option takes the form of a sequence of value seperated
by the '|' character. The values have the following significance.
"OwnDB" | "SetPeerAddress" | "DNSReverseLookup"
<CENTER>
<STRONG>Flags</STRONG>
<TABLE BORDER=1>
<TH>Value
<TH>Meaning

<TR>
<TD>OwnDB
<TD>Each child object will create its own DB

<TR>
<TD>SetPeerAddress
<TD>On creation, each child socket will get the IP address of its remote
peer and place it in the DB with the variable name specified with the
'PeerAddressVariable' option.

<TR>
<TD>DNSReverseLookup
<TD>When the IP address of the remote peer is retrieved the hostname of
the peer will be found and placed in the DB with the variable name
specified by the 'PeerHostNameVariable' option.

<TR>
<TD>SetLocalAddress
<TD>On creation, each child socket will store its local IP address in
the variable specified by the 'LocalAddressVariable' parameter.

</TABLE>
</CENTER>

Returns:
	PIAPI_COMPLETED on success.
	PIAPI_CONTINUE if no action taken.
	PIAPI_ERROR and PIAPI_ABORT respectively for generic and severe
	error conditions.

Example:
	<PRE>

	&ltObject&gt
		Name DefaultUDPIPIOObject
		Class UDPIPIOClass
		BindHost localhost
		BindPort 53
		Type Passive
		AcceptTimeout -1
		RecvTimeout 30
		SendTimeout 60 
		PeerAddressVariable "STRING:RemoteAddr"
		PeerHostNameVariable "STRING:RemoteHost"
		LocalAddressVariable "STRING:LocalAddr"
		Flags "OwnDB|SetLocalAddress|SetPeerAddress|DNSReverseLookup"
	&lt/Object&gt

	</PRE>
/*___+++HTMLDOC_END+++___*/
#endif

/*____________________________________________________________________________*\
** Data structure associated with the instances of prototype UDPIPSocket object
\*____________________________________________________________________________*/
class MainSocketData
{
private:
	PIDB *pDB;
	PIConfig *pConfigDB;

	/* --- internal methods --- */
	int FlagNamesToNumber( const char *pFlags )
		{
		assert( pFlags );
		if ( !pFlags )
			{ return 0; };
		int iFlags = 0;
		for(;;)
			{
			/* --- skip leading space --- */
			for( ; *pFlags && (isspace(*pFlags)); pFlags++ );
			const char *pFlag = pFlags;
			int i=0;
			for( ; pFlags[i] && !(isspace(pFlags[i])) && pFlags[i]!='|'; i++ );
			
			/* --- deal with this flag --- */
			if ( !PIUtil_strncmpi( pFlag, VALUE_FLAG_OWNDB, i ))
				{ iFlags |= FLG_OWNDB; }
			else if ( !PIUtil_strncmpi( pFlag, VALUE_FLAG_SETREMOTEADDRESS, i ))
				{ iFlags |= FLG_SETREMOTEADDRESS; }
			else if ( !PIUtil_strncmpi( pFlag, VALUE_FLAG_DNSREVERSELOOKUP, i ))
				{ iFlags |= FLG_DNSREVERSELOOKUP; }
			else if ( !PIUtil_strncmpi( pFlag, VALUE_FLAG_SETLOCALADDRESS, i ))
				{ iFlags |= FLG_SETLOCALADDRESS; }
			else
				{ iFlags |= FLG_ERROR; };

			/* -- check for end of string --- */
			if ( pFlags[i] )
				{
				/* --- another flag --- */
				pFlags=&(pFlags[++i]);
				}
			else
				{ break; };
			};
		return iFlags;
		};

public:
	int iServer;				/* active or passive, 0 = passive */
	int iPort;
	char *pHost;
	int iAcceptTimeout;
	int iRecvTimeout;
	int iSendTimeout;
	DBInfo *pPeerAddressVariable;	/* variable used to store remote IP */
	DBInfo *pPeerHostVariable;		/* variable used to store remote Host */ 
	DBInfo *pLocalAddressVariable;	/* variable used to store local IP */ 
	int iFlags;
	char *pFAcceptLockFile;		/* lockfile name */
	int iFAcceptLockFD;			/* lockfile descriptor */

protected:
	int Parameter( const char *pVariable, const char *pValue,
		const char *pWhere )
		{
		assert( pVariable && pValue );
		if ( !pVariable || !pValue )
			{ return 0; };
		if ( !PIUtil_stricmp( KEY_CONF_TYPE, pVariable ) )
			{
			iServer = !PIUtil_stricmp( pValue, VALUE_PASSIVE );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_BINDPORT, pVariable ) )
			{
			iPort = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_BINDHOST, pVariable ) )
			{
			pHost = PIUtil_strdup( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_ACCEPTTIMEOUT, pVariable ) )
			{
			iAcceptTimeout = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_SENDTIMEOUT, pVariable ) )
			{
			iSendTimeout = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_RECVTIMEOUT, pVariable ) )
			{
			iRecvTimeout = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_ACCEPTTIMEOUT, pVariable ) )
			{
			iAcceptTimeout = atoi( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_PEERADDRESSVARIABLE, pVariable ) )
			{
			pPeerAddressVariable = new DBInfo( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_PEERHOSTNAMEVARIABLE, pVariable ) )
			{
			pPeerHostVariable = new DBInfo( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_LOCALADDRESSVARIABLE, pVariable ) )
			{
			pLocalAddressVariable = new DBInfo( pValue );
			}
		else if ( !PIUtil_stricmp( KEY_CONF_ACCEPTLOCKFILE, pVariable ) )
			{
/* --- lockfile synchronization is UNIX specific --- */
#if POSIX
			/* --- check this hasn't already been done --- */
			if ( iFAcceptLockFD!=-1 )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"%sUDPIPIO: 'AcceptLockFile' specified more than once.",
					pWhere );
				return 0;
				};
			
			/* --- get the name of the lockfile --- */	
			/* NOTE: There is some risk of overwrite here if file specified by
			** UNIXDaemon is too long. Fortunately this is under our control
			** not under a remote users control.
			*/
			enum { BUF_SIZE=511 };
			char szBuf[BUF_SIZE+1];
			*szBuf = '\0';
			ifstream ifs( pValue );
			if ( ifs.bad() )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"%sUDPIPIO: Could not open file: '%s'.%s",
					pWhere, pValue, szLockFileMsg );
				return 0;
				};
			if ( ifs.good() )
				{
				ifs >> szBuf;
				};	
			if ( !*szBuf )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"%sUDPIPIO: Could not read lock file name from file:\
 '%s'.", pWhere, pValue );
				return 0;
				};

			pFAcceptLockFile = PIUtil_strdup( szBuf );

			/* --- attempt to open a descriptor to the lockfile --- */
			iFAcceptLockFD = ::open( pFAcceptLockFile, O_RDWR );
			if ( iFAcceptLockFD==-1 )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"UDPIPIO%s: Error opening lockfile: '%s' errno: %d.",
					pWhere, pFAcceptLockFile, errno );

				return 0;
				};
#endif	/* lockfile synchronization is UNIX specific */
			}
		else if ( !PIUtil_stricmp( KEY_CONF_FLAGS, pVariable ) )
			{
			iFlags |= FlagNamesToNumber( pValue );
			if ( iFlags & FLG_ERROR )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"%sUDPIPIO: Bad flag in list of flags '%s'", pWhere,
					pValue );
				return 0;
				};
			}
		else
			{
			PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
				"%sUDPIPIO: Unknown parameter '%s'", pWhere, pVariable );
			return 0;
			};

		return 1;
		};

public:
	static int ParameterFn( void *pData, const char *pVar, const char *pVal,
		const char *pWhere )
		{
		assert( pData );
		return ((MainSocketData *)pData)->Parameter( pVar, pVal, pWhere );
		};

	/* ---
	This pointer is here because a thread could be terminated just when it
	was waiting on a connection to be accepted. The new IO object
	would not be deleted because the pointer has not been returned
	to whomever would use it. Therefore the main IO object sets this
	pointer to a new IO object just before polling, and resets it
	to 0 just after, so that if the thread and main parent object
	is deleted without the poll returning, then the new IO object would
	still be deleted.
	--- */
	PIObject *pAcceptingIOObject;
	
	MainSocketData( PIDB *pTheDB, PIConfig *pTheConfigDB )
		:
			pDB( pTheDB ),
			pConfigDB( pTheConfigDB ),
			iServer( 0 ),
			iPort( 0 ),
			pHost( 0 ),
			iAcceptTimeout( DEFAULT_TIMEOUT ),
			iRecvTimeout( DEFAULT_TIMEOUT ),
			iSendTimeout( DEFAULT_TIMEOUT ),
			pPeerAddressVariable( 0 ),
			pPeerHostVariable( 0 ),
			pLocalAddressVariable( 0 ),
			iFlags( 0 ),
			pFAcceptLockFile( 0 ),
			iFAcceptLockFD( -1 ),
			pAcceptingIOObject( 0 )
		{};

	~MainSocketData()
		{
		PIUtil_free( pHost ); 
		delete pPeerAddressVariable; 
		delete pPeerHostVariable; 
		delete pLocalAddressVariable; 
		PIUtil_free( pFAcceptLockFile ); 

		/* --- close lockfile descriptor --- */
#	if POSIX
		if ( iFAcceptLockFD!=-1 )
			{ ::close( iFAcceptLockFD ); };
		iFAcceptLockFD = -1;
#	endif

		if ( pAcceptingIOObject )
			{ PIObject_delete( pAcceptingIOObject, 0, 0 ); };
		};

	int Accept( UDPIPSocketData *pData, PIObject *pChildSocket );
	int InitSocketFromConfiguration( PIObject *pThis,
		UDPIPSocketData *pProtoType );
};

/*____________________________________________________________________________*\
** Data structure associated with the instances of UDPIPSocket objects
\*____________________________________________________________________________*/
class UDPIPSocketData
{
public:
	/* --- generic stuff --- */
	PIObject *pObject;
	int iPrototype;
	int iOK;

	/* --- socket stuff --- */
	int iSize;
	struct sockaddr_in tAddr;
	int iSocket;

    /* --- buffer for received data --- */
    int iReceivedLen;
    char szReceived[MAX_PACKET_SIZE];

	/* --- other stuff --- */
	const UDPIPSocketData *pParentData;		/* NULL for parent sockets */
	MainSocketData *pMainSocketData;		/* NULL for child sockets */

	/* ---
	Constructor
	--- */
	UDPIPSocketData(
		PIObject *pTheObject,
		int iArgc,
		const char *ppArgv[] )
	:
		pObject( pTheObject ),
		iPrototype( 1 ),	
		iOK( 0 ),
		iSize( sizeof( struct sockaddr_in ) ),
		iSocket( SOCKET_ERROR ),
        iReceivedLen( 0 ),
		pParentData( 0 ),
		pMainSocketData( 0 )
		{
		PIDB *pTheDB;
		PIConfig *pDB;
		MainSocketData *pConfig;

		pTheDB = PIObject_getDB( pObject );
		pDB = PIObject_getConfigurationDB( pObject );
		if ( !pTheDB || !pDB )
			{ return; };
		pConfig = new MainSocketData( pTheDB, pDB );
		if (  !pConfig )
			{ return; };

		if ( !PIObject_readParameters( pObject, iArgc, ppArgv,
			MainSocketData::ParameterFn, pConfig ) )
			{ return; };

		if ( pConfig->InitSocketFromConfiguration( pTheObject, this ) )
			{ return; };
		iOK = 1;
		};

	/* ---
	Copy constructor 
	--- */
	UDPIPSocketData(
		PIObject *pTheObject,
		const UDPIPSocketData *pTheParentData,
		int /* iArgc */,
		const char * /* ppArgv */ [] )
	:
		pObject( pTheObject ),
		iPrototype( 0 ),	
		iOK( 0 ),
		iSize( sizeof( struct sockaddr_in ) ),
		iSocket( SOCKET_ERROR ),
		pParentData( pTheParentData ),
		pMainSocketData( pTheParentData->pMainSocketData )
		{
		PIObject_setUserData( pTheObject, this );
		assert( pMainSocketData );
		
		PIDB *pParentDB = PIObject_getDB( pObject );
		PIDB *pDB = pParentDB;

		/* --- should this object have its own DB ? --- */
		if ( pMainSocketData->iFlags & FLG_OWNDB )
			{
			pDB = PIDB_new( 0, "UDPIPSocketDB" );
			PIObject_setDB( pObject, pDB );
			};

		iOK = 1;
		};

	~UDPIPSocketData()
		{
		if ( iSocket!=-1 )
			{
			::shutdown( iSocket, 2 );
			::closesocket( iSocket );
			};
		if ( iPrototype )
			{
			delete pMainSocketData;
			}
		else
			{
			PIDB_delete( PIObject_getDB( pObject ) );
			};

		/* ---
		this object does not have to delete its DB object even
		if it created it, because that happens automatically
		--- */
		};

	inline int IsOK()		{ return iOK; };
};

#define CONFIG_ERR(pThis, msg)	\
	{ PILog_addMessage( PIObject_getDB(pThis), \
	PIObject_getConfigurationDB(pThis), PILOG_ERROR, (msg) ); }

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int MainSocketData::InitSocketFromConfiguration( PIObject *pThis,
	UDPIPSocketData *pProtoTypeUDPIPObject )
{
	PIDB *pTheDB;
	PIConfig *pDB;

	pTheDB = PIObject_getDB( pThis );
	pDB = PIObject_getConfigurationDB( pThis );
	if ( !pTheDB || !pDB )
		{ return PIAPI_ABORT; };

	/*
	**
	*/
	if ( !iServer )
		{
		pProtoTypeUDPIPObject->pMainSocketData = this;
		return Internal_initActiveSocket( pThis );
		};

	/*
	** Check for errors
	*/
	if ( iPort<0 )
		{
		PILog_addMessage( pTheDB, pDB, PILOG_ERROR,	
			"UDPIPIO: Invalid bind port." );
		}
	else if ( !pHost )
		{
		PILog_addMessage( pTheDB, pDB, PILOG_ERROR,
			"UDPIPIO: Invalid bind host." );
		}
	else
		{
		/* --- OK --- */
		pProtoTypeUDPIPObject->pMainSocketData = this;

		if ( iServer )
			{
			return Internal_initPassiveSocket( pThis, pProtoTypeUDPIPObject );
			};
		};
	
	/* --- something failed --- */
	return PIAPI_ABORT;
}

/*____________________________________________________________________________*\
 *
 Function:		 Internal_initActiveSocket
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_initActiveSocket( PIObject * )
{
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 Internal_initPassiveSocket
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int Internal_initPassiveSocket( PIObject *pThis, UDPIPSocketData *pData )
{
	/* automatic variables */
	int iOne=1;

	assert( pData );

	/*
	** Make the socket 
	*/
	pData->iSocket = ::socket( AF_INET, SOCK_DGRAM, 0);
	if ( pData->iSocket<0 )
		{
		CONFIG_ERR(pThis, "socket() failed" );
		return PIAPI_ERROR;
		};

	/*
	**
	** Attempt to bind to port
 	**
	*/ 
	pData->iSize=sizeof( struct sockaddr_in );
	pData->tAddr.sin_family = AF_INET;
/**************/

#if 0
	/*
	** Sequence for binding to a specific address
	*/
	/* --- get local hostname and hostent --- */
	enum { BUF_SIZE=256 };
	char szBuf[BUF_SIZE+1];

	if ( ::gethostname( szBuf, BUF_SIZE)==-1 )
		{
		CONFIG_ERR(pThis, "gethostname() failed" );
		return PIAPI_ERROR;
		};

	struct hostent *phe = ::gethostbyname( szBuf );
	if ( !phe )
		{
		CONFIG_ERR(pThis, "gethostbyname() failed" );
		return PIAPI_ERROR;
		};

	/* --- setup local address structure -- */
	memcpy( &(pData->tAddr.sin_addr.s_addr), phe->h_addr, phe->h_length );
#else
	/*
	** Sequence for binding to any address
	*/
	pData->tAddr.sin_addr.s_addr=htonl(INADDR_ANY);
#endif
/**************/
	pData->tAddr.sin_port = htons( pData->pMainSocketData->iPort );

#if 0
// LATER
	/*
 	** Allow host,port to be reused (or attempt to reuse existing host,port)
	*/
	if ( ::setsockopt( pData->iSocket, SOL_SOCKET, SO_REUSEADDR, 
		(const char *)&iOne, sizeof( iOne ) )!=0 )
		{
		CONFIG_ERR(pThis, "setsockopt() failed for SO_REUSEADDR" );
		return PIAPI_ERROR;
		};
#endif

	if ( ::bind( pData->iSocket, (struct sockaddr *)&(pData->tAddr), 
		pData->iSize)<0) 
		{
		const char *pError = "";
		switch( errno )
			{	
			case EACCES:
				pError = szBindFailedAccess;
				break;

			case EINVAL:
#if POSIX
			case EADDRINUSE:
#endif
				pError = szBindFailedInUse;
				break;

			default:;
			};

		PILog_addMessage(
			PIObject_getDB( pThis ),
			PIObject_getConfigurationDB( pThis ), PILOG_ERROR,
			"Failed to bind to port %d.%s", pData->pMainSocketData->iPort,
			pError );

		return PIAPI_ERROR;
		};
		
	/*
	** Done
	*/
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		Internal_getHostEnt
 Synopsis:		private
 Description:
	returns a struct hostent *, given either an id address or hostname
	as parameter.
\*____________________________________________________________________________*/
struct hostent *Internal_getHostEnt( const char *pHost )
{
	int i;
	int iIsAddr=1;
	int iLen;
	struct hostent *phe;

	/*
	** If the string consists only of decimal digits and '.'
	** assume its an IP address
	*/
	if ( !pHost ) return 0;
	iLen = strlen( pHost );
	for(i=0; i<iLen; i++)
		{
		if ( !isdigit( pHost[i] ) && pHost[i]!='.' )
			{ 		
			iIsAddr=0; 
			break; 
			};
		};

	/* --- get hostent * --- */
	phe=0;
	if ( iIsAddr )
		{
		struct in_addr tAddr;
		u_long ul;
		int iSize;

		ul = inet_addr( pHost );
		iSize = sizeof( struct in_addr );
		memcpy( &tAddr, &ul, iSize );
		phe = gethostbyaddr( (const char *)&tAddr, iSize, AF_INET);
		}
	else
		{
		phe = gethostbyname( pHost );
		};
	return phe;
}

/*____________________________________________________________________________*\
 *
 Function:		 Internal_connect
 Synopsis:
 Description:
	Open a connection to a remote host
\*____________________________________________________________________________*/
static int Internal_connect( PIObject *pThis,
		const char *pHostName, int iPort )
{
	/* automatic variables */
	UDPIPSocketData *pData=(UDPIPSocketData *)PIObject_getUserData( pThis );
	struct hostent *phe;

	assert( pData );

	/*
	** Make the socket
	*/
	pData->iSocket = ::socket( AF_INET, SOCK_DGRAM, 0);
	if ( pData->iSocket==SOCKET_ERROR )
		{
		PIPlatform_setLastError( PIAPI_EXHAUSTED );
		CONFIG_ERR(pThis, "socket() failed" );
		return PIAPI_ERROR;
		};

	/*
	** setup pData->tAddr with remote host information.
	*/
	phe=Internal_getHostEnt( pHostName );
	if ( !phe )
		{
		PIPlatform_setLastError( PIAPI_ERROR );
		return PIAPI_ERROR;
		};

	pData->iSize=sizeof( struct sockaddr_in );
	pData->tAddr.sin_family = AF_INET;
	memcpy( &(pData->tAddr.sin_addr), phe->h_addr, phe->h_length );
	pData->tAddr.sin_port = htons( iPort );

	/*
 	** bind 
	*/
	/* or not to bind */

	/*
 	** connect
	*/
	if ( connect( pData->iSocket, (struct sockaddr *)&(pData->tAddr), 
			pData->iSize )<0 )
		{
		PIPlatform_setLastError( PIAPI_ERROR );
		return PIAPI_ERROR;
		};

	/* ---
	Write send and receive timeout into the DB, to be overridden as
	necessary
	--- */
	assert( pFKRecvTimeout );
	assert( pFKSendTimeout );
	PIDB_add( PIObject_getDB( pThis ), PIDBTYPE_OPAQUE, pFKRecvTimeout,
		(void *)(pData->pMainSocketData->iRecvTimeout), PIDBFLAG_FASTKEY );
	PIDB_add( PIObject_getDB( pThis ), PIDBTYPE_OPAQUE, pFKSendTimeout,
		(void *)(pData->pMainSocketData->iSendTimeout), PIDBFLAG_FASTKEY );

	/*
	** Done
	*/
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:
 Synopsis:
 Description:
\*____________________________________________________________________________*/
int MainSocketData::Accept( UDPIPSocketData *pData, PIObject *pNewSocket )
{
	/* --- automatic variables --- */
    int iRet;
	UDPIPSocketData *pChildData;

	if ( !pNewSocket ) return PIAPI_ABORT;

	/* --- get data object of child class --- */
	pChildData=(UDPIPSocketData *)PIObject_getUserData( pNewSocket );

	/*
	** Accept a new connection
	*/
	iRet=-1;

	/* --- note the new accepting IO object, so it can be deleted --- */
	pAcceptingIOObject = pNewSocket;

	for(;;)
		{		

		/* --- test that accept is possible --- */
		if ( PIPlatform_pollNetFD( pData->iSocket, PIPLATFORM_POLL_READ,
			iAcceptTimeout ) <=0 )
			{
			break;
			};

#if POSIX
		/* ---
		wait to acquire the POSIX lock to stop multiple 	
		processes colliding on the accept 
		--- */
		if ( iFAcceptLockFD!=-1 )
			{
			/* ---
			Lock
			--- */
			if ( PIPlatform_advisaryLockFd( iFAcceptLockFD )!=PIAPI_COMPLETED )
				{
				/* --- an error occurred in lock --- */
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"UDPIPIO: Error obtaining lock (locking failed) errno: \
'%d'.\
 Removing lockfile to force server shutdown.", errno );

				/* ---
				remove lockfile to shutdown entire server
				see notes above.
				--- */
				::unlink( pFAcceptLockFile );	/* --- remove file --- */
				return PIAPI_ABORT;
				};
			};

#endif

#if 0
		/* --- accept the connection --- */
		iNewS=::accept( pData->iSocket, 
				(struct sockaddr *)&(pChildData->tAddr),
				&(pChildData->iSize) );
#endif
        pChildData->iSocket = pData->iSocket;
		iRet = ::recvfrom(
            pData->iSocket,
            pChildData->szReceived,
            MAX_PACKET_SIZE,
            0,
            (struct sockaddr *)&(pChildData->tAddr),
            &(pChildData->iSize) );

		if ( iFAcceptLockFD!=-1 )
			{
			/* ---
			Release the lock
			--- */
			if ( PIPlatform_advisaryUnlockFd( iFAcceptLockFD )!=PIAPI_COMPLETED )
				{
				PILog_addMessage( pDB, pConfigDB, PILOG_ERROR,
					"UDPIPIO: Error releasing lock, (unlock failed) errno:\
'%d'.", errno );
				::unlink( pFAcceptLockFile );	/* --- remove file --- */
				return PIAPI_ABORT;
				};
			};

		/* ---
		Some cases of error on accept are restartable, don't return an error
		for those
		--- */
		if ( iRet==-1 )
			{
#if POSIX
			if ( errno==EINTR )	{ /* just a signal */ continue; };
			if ( errno==ECONNRESET ) { /* aborted connection */ continue; };
#endif
			};

		/* --- done --- */
		break;
		};

	/* --- NULL out the pointer to pNewSocket --- */
	pAcceptingIOObject = 0;

	if ( iRet<0 ) 
		{
		return PIAPI_ERROR; /* non-critical error, callee may try again */
		};

    pChildData->iReceivedLen = iRet;

	/* ---
	**
	**
	** This is a new connection. Look at the flags and variable names
	** to see what lookups should occur and where the data is put
	**
	**
	--- */

	/* --- should this object set the PeerIP address? --- */
	int iFlags = pChildData->pMainSocketData->iFlags;

	/* ---
	Set PeerAddress variable, if applicable
	--- */
	/* ---
	When the hostname is unavailable jsut have it turn up blank
#	define TEXT_UNAVAILABLE	"** Unavailable **"
	--- */
#	define TEXT_UNAVAILABLE	""
	if ( iFlags & FLG_SETREMOTEADDRESS )
		{
		/* --- set remote Address ? --- */
		DBInfo *pIPInfo = pChildData->pMainSocketData->pPeerAddressVariable;
		if ( pIPInfo )
			{
			/* ---
			inet_ntoa is not thread safe, protect access to it
			--- */
			PIPlatform_beforeUnsafeBlock();

			/* --- convert the peer ip to a string if we have the ip --- */
			const char *pPeerAddress = 0;
			pPeerAddress = ::inet_ntoa( pChildData->tAddr.sin_addr );

			PIDB_add( PIObject_getDB( pNewSocket ), pIPInfo->iType,
				pIPInfo->pFKVariableName, (void *)pPeerAddress,
				pIPInfo->iFlags | PIDBFLAG_FASTKEY );

			PIPlatform_afterUnsafeBlock();
			};

		if ( iFlags & FLG_DNSREVERSELOOKUP )
			{
			/* --- set remote hostname ? --- */
			DBInfo *pHostInfo = pChildData->pMainSocketData->pPeerHostVariable;
			if ( pHostInfo )
				{
				/* ---
				inet_ntoa is not thread safe, protect access to it
				--- */
				PIPlatform_beforeUnsafeBlock();

				/* --- if we have the ip structure, lookup the hostname --- */
				struct hostent *pEnt;
				pEnt = ::gethostbyaddr(
					(const char *)&(pChildData->tAddr.sin_addr),
					sizeof(pChildData->tAddr.sin_addr), AF_INET ); 
				PIDB_add( PIObject_getDB( pNewSocket ), pHostInfo->iType,
					pHostInfo->pFKVariableName,
					( pEnt ? (void *)pEnt->h_name : TEXT_UNAVAILABLE ),
					pHostInfo->iFlags | PIDBFLAG_FASTKEY ); 
				PIPlatform_afterUnsafeBlock();
				};
			};
		};

	/* ---
	Set LocalAddress if applicable
	--- */
	if ( iFlags & FLG_SETLOCALADDRESS )
		{
		DBInfo *pIPInfo = pChildData->pMainSocketData->pLocalAddressVariable;
		if ( pIPInfo )
			{
			struct sockaddr_in tLocalAddr;
			int iLocalSize = sizeof( struct sockaddr_in );

			for(;;)
				{
				/* --- get local socket information --- */
				if ( ::getsockname( pChildData->iSocket,
						(struct sockaddr *)&tLocalAddr,
						&iLocalSize )==-1 )
					{
					if ( errno==EINTR )
						{
						/* --- restart system call --- */
						PIThread_userYield();
						continue;
						};
					iLocalSize = 0;	/* indicate an error */
					};
				break;
				};

			if ( iLocalSize>0 )
				{
				/* ---
				inet_ntoa is not thread safe, protect access to it
				--- */
				PIPlatform_beforeUnsafeBlock();
	
				const char *pLocalAddress = 0;
	
				pLocalAddress = ::inet_ntoa( tLocalAddr.sin_addr );
				PIDB_add( PIObject_getDB( pNewSocket ), pIPInfo->iType,
					pIPInfo->pFKVariableName, (void *)pLocalAddress,
					pIPInfo->iFlags | PIDBFLAG_FASTKEY );

				PIPlatform_afterUnsafeBlock();
				}
			else
				{
				/* --- error --- */
				STDERR << "getsockname() failed, errno#" << errno << endl;
				PIDB_add( PIObject_getDB( pNewSocket ), pIPInfo->iType,
					pIPInfo->pFKVariableName, (void *)"0.0.0.0",
					pIPInfo->iFlags | PIDBFLAG_FASTKEY );
				};	/* iLocalSize!=0 */

			};	/* pIPInfo!=NULL */
		};	/* iFlags & FLG_SETLOCALADDRRESS */

	/* ---
	Write send and receive timeout into the DB, to be overridden as
	necessary
	--- */
	assert( pFKRecvTimeout );
	assert( pFKSendTimeout );
	PIDB_add( PIObject_getDB( pNewSocket ), PIDBTYPE_OPAQUE, pFKRecvTimeout,
		(void *)(pChildData->pMainSocketData->iRecvTimeout), PIDBFLAG_FASTKEY );
	PIDB_add( PIObject_getDB( pNewSocket ), PIDBTYPE_OPAQUE, pFKSendTimeout,
		(void *)(pChildData->pMainSocketData->iSendTimeout), PIDBFLAG_FASTKEY );

	/* ---
	Done
	--- */
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 UDPIPSocket_onClassLoad
 Synopsis:
 Description:
	This is called on program startup, restart or shutdown.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_onClassLoad( PIClass_LoadAction eAction,
	void * /* pV */ )
{
#if WIN32
	WSAData tWSAData;
	enum { DESIRED_WINSOCK_VERSION=0x0101 };
	/* enum { DESIRED_WINSOCK_VERSION=0x0200 }; */
#endif

	switch( eAction )
		{
		case STARTUP:
#if WIN32
			/*
			** WSAStartup returns 0 or one of these values
			**
			WSASYSNOTREADY
			WSAVERNOTSUPPORTED
			WSAEINVAL
			*/
			if ( ::WSAStartup( DESIRED_WINSOCK_VERSION, &tWSAData ) )
				{ return PIAPI_ABORT; };
#endif

			/* ---
			Get fast keys
			--- */
			pFKSendTimeout = PIDB_getFastKey( KEY_CONF_SENDTIMEOUT, 
				PIDBTYPE_OPAQUE );
			pFKRecvTimeout = PIDB_getFastKey( KEY_CONF_RECVTIMEOUT, 
				PIDBTYPE_OPAQUE );
			break;

		case SHUTDOWN:
#if WIN32
			WSACleanup();
#endif
			break;
		
		default:;
		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 UDPIPSocket_UDPIPSocket
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_UDPIPSocket(
	PIObject *pObject,
	int iArgc,
	const char *ppArgv[]
	)
{
	/*
	** Allocate instance data
	*/
	UDPIPSocketData *pData = new UDPIPSocketData( pObject, iArgc, ppArgv );	
	PIObject_setUserData( pObject, pData );

	/*
	*/
	/* --- pObject is the new parent socket --- */
	if ( !pData->IsOK() )
		{ return PIAPI_ABORT; };

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		 UDPIPSocket_copyUDPIPSocket
 Synopsis:
 Description:
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_copyUDPIPSocket( PIObject *pObject, int iArgc,
	const char *ppArgv[] )
{
	/* automatic variables */
	UDPIPSocketData *pData;
	UDPIPSocketData *pParentData;

	/*
	** Grap old user data, as pParentData
	*/
	pParentData = (UDPIPSocketData *)PIObject_getUserData( pObject ); 

	/*
	** Allocate instance data
	*/
	pData = new UDPIPSocketData( pObject, pParentData, iArgc, ppArgv );

	/* ---
	Is this an active or passive socket?
	--- */
	if ( pData->pMainSocketData->iServer )
		{
		/* --- 
		Attempt to accept a new connection on pObject socket
		--- */
		int iRet = pData->pMainSocketData->Accept( pParentData, pObject );
		if ( iRet!=PIAPI_COMPLETED )
			{ return iRet; };
		}
	else
		{
		/* --- 
		Attempt to make a new connection to a remote host
		--- */
		if ( iArgc!=2 )
			{
			PIPlatform_setLastError( PIAPI_EINVAL );
			return PIAPI_ERROR;
			};

		if ( Internal_connect( pObject, (const char *)ppArgv[0],
			(int)ppArgv[1] ) )
			{
			return PIAPI_ERROR;
			};

		};

	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:		UDPIPSocket_xUDPIPSocket
 Synopsis:
 Description:
	Destructor for instances of the UDPIPSocket class. Performs cleanup.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_xUDPIPSocket( PIObject *pObject, int,
	const char *[] )
{
	delete (UDPIPSocketData *)PIObject_getUserData( pObject );
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:	   UDPIPSocket_send
 Synopsis:
 Description:
	Send data down pObject socket.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_send( PIObject *pObject, void *pData,
	int iLen, int *piSent, int, const char *[] )
{
	/* --- automatic variables --- */
	int iRet = -1;
	int iToSend=iLen;
	int iSent=0;
	const char *pToSend=(const char *)pData;
	UDPIPSocketData *pInstanceData=
	(UDPIPSocketData *)PIObject_getUserData( pObject );
	assert( pToSend && pInstanceData && iToSend && piSent );

	int iTimeout=(int)PIDB_lookup( PIObject_getDB( pObject ),
		PIDBTYPE_OPAQUE, pFKSendTimeout, PIDBFLAG_FASTKEY );
	do	{	
		if ( PIPlatform_pollNetFD( pInstanceData->iSocket,
			PIPLATFORM_POLL_WRITE, 
			iTimeout ) <=0 )
			{ break; };
		iRet=::sendto(
            pInstanceData->iSocket,
            &(pToSend[iSent]),
            iToSend,
            0,
            (struct sockaddr *)&(pInstanceData->tAddr),
            pInstanceData->iSize ); 
		if ( !iRet ) break;							/* connection closed */
		if ( iRet==-1 )
			{
			if ( errno==EINTR )
				{ continue;	/* signal caught */ };

			break;
			};
		iSent += iRet;
		iToSend -= iRet;
		} 
	while( iToSend>0 ); 
	
	if ( iRet==-1 )
		{ return PIAPI_ERROR; };
	*piSent=iSent;
	return PIAPI_COMPLETED;
}

/*____________________________________________________________________________*\
 *
 Function:	   UDPIPSocket_recv
 Synopsis:
 Description:
	Get data out of pObject socket.
\*____________________________________________________________________________*/
PUBLIC_PIAPI int UDPIPSocket_recv( PIObject *pObject, void *pData,
	int iLen, int *piReceived, int, const char *[] )
{
	/* --- automatic variables --- */
	int iRet = -1;
	UDPIPSocketData *pInstanceData=
		(UDPIPSocketData *)PIObject_getUserData( pObject );
	assert( pInstanceData && piReceived );

    /*
    ** Check for received data in buffer
    */
    if ( pInstanceData->iReceivedLen > 0 )
        {
        if ( iLen < pInstanceData->iReceivedLen )
            {
            *piReceived = iLen;
            memcpy( pData, pInstanceData->szReceived, iLen );
            pInstanceData->iReceivedLen -= iLen;
            }
        else
            {
            *piReceived = pInstanceData->iReceivedLen;
            memcpy( pData, pInstanceData->szReceived,
                pInstanceData->iReceivedLen );
            pInstanceData->iReceivedLen = 0;
            };

        return PIAPI_COMPLETED;
        };
    
    /*
    ** Read more data
    */
	int iTimeout=(int)PIDB_lookup( PIObject_getDB( pObject ),
		PIDBTYPE_OPAQUE, pFKRecvTimeout, PIDBFLAG_FASTKEY );
	for(;;)
		{	
		if ( PIPlatform_pollNetFD( pInstanceData->iSocket,
			PIPLATFORM_POLL_READ, 
			iTimeout ) <=0 )
			{ break; };

		iRet=::recvfrom(
            pInstanceData->iSocket,
            (char *)pData,
            iLen,
            0,
            (struct sockaddr *)&(pInstanceData->tAddr),
            &(pInstanceData->iSize) );

		if ( iRet==-1 && errno==EINTR ) continue;	/* signal caught */
		break;
		}; 
	
	if ( iRet==-1 ) return PIAPI_ERROR;
	*piReceived=iRet;
	return PIAPI_COMPLETED;
}

/* ---
	Class configuration specification 
--- */
#if 0
/*___+++CNF_BEGIN+++___*/
	<Class>
		Name UDPIPIOClass
		Type IO
		Library IO
		OnClassLoad UDPIPSocket_onClassLoad
		Constructor UDPIPSocket_UDPIPSocket
		CopyConstructor UDPIPSocket_copyUDPIPSocket
		Destructor UDPIPSocket_xUDPIPSocket
		Out UDPIPSocket_send
		In UDPIPSocket_recv
	</Class>

/*___+++CNF_END+++___*/

	<Object>
		Name UDPIPIOObject
		Class UDPIPIOClass
		BindHost localhost			# host to bind to
		BindPort 1200				# port to bind on 
		Type Passive				# Passive (server) or Active (client)
		AcceptTimeout 1000			# applies to Passive only
		RecvTimeout 30
		SendTimeout 60 
		PeerAddressVariable "STRING:RemoteAddr"
		PeerHostNameVariable "STRING:RemoteHost"
		LocalAddressVariable "STRING:LocalAddr"
		Flags "OwnDB|SetLocalAddress|SetPeerAddress|DNSReverseLookup"
	</Object>

#endif

